% This script reproduces Figures 5.1 and 5.2 in [1]
%
% [1] I. Dassios, K. Fountoulakis and J. Gondzio, "A Preconditioner for a
%     Primal-Dual Newton Conjugate Gradients Method for Compressed Sensing 
%     Problems". Technical Report ERGO 14-021.
%
% ** Signal Processing Toolbox is required **
%
% This file should take about 3.412300e+02 sec. CPU time to be completed.

clear;
tt = cputime;

randn('state',245);
rand('state',245);

% Generate phantom image.
psize = 512;

n1 = psize; 
n2 = n1;
N = n1*n2;
n = max(n1,n2);

x = phantom(n);

x = x(1:n1,1:n2);
x_original = x;
mat = @(x) reshape(x,n1,n2);
vec = @(x) x(:);

% Partial Fourier measurements.
M = round( N / 4 );
omega = randn(N,1);
[temp,omega] = sort(omega);
omega = sort( omega(1:M) );
downsample = @(x) x(omega);
SS.type = '()'; SS.subs{1} = omega; SS.subs{2} = ':';
upsample = @(x) subsasgn( zeros(N,size(x,2)),SS,x);

% take partial 2D DCT measurements.  We have Af(At) = identity (not
% vice-versa of course). Randomly permuted.
rp = randperm(N);
[~,rp_inv] = sort(rp);
rpF = @(x) x(rp);
rp_invF = @(x) x(rp_inv);
        
my_dct2 = @(x) dct(dct(x).').';
my_idct2= @(x) idct(idct(x).').';
Af = @(x) downsample(vec( my_dct2( mat(rpF(mat(x) )) ) ) );
At = @(y) vec( rp_invF( my_idct2( mat( upsample( y ) ) ) ) );


% Make matrix A.
A = linop_handles([M,N], Af, At );

% Noise function.
myAwgn = @(x,snr) x + ...
        10^( (10*log10(sum(abs(x(:)).^2)/length(x(:))) - snr)/20 )*randn(size(x));
    
% Generate samples.
b_original = Af(vec(x_original));
snr = 10;  % SNR in dB
x_noisy = myAwgn( x_original, snr );
b   = Af( vec(x_noisy) );

x_ref   = x_original;

sz              = A([],0);
M               = sz{2}(1);
N               = sz{1}(1);
[n1,n2]         = size(x_original);

fprintf('\tA is %s, %d x %d, signal has SNR %d dB\n', 'partial Fourier', M, N, round(snr) );

%% Call pdNCG solver for Figures 5.1a and 5.1b.
Apdncg  = A_operator(@(w) A(w,1), @(w) A(w,2));
global counter;
global fig1a;
global fig1b;
global cgiter_prec;
global cpu_prec;
cgiter_prec = zeros(N,1);
cpu_prec = zeros(N,1);
global cgiter_unprec;
global cpu_unprec;
cgiter_unprec = zeros(N,1);
cpu_unprec = zeros(N,1);
counter = 1;

opts.n1  = n;
opts.n2  = n;
opts.b   = b;
opts.A   = Apdncg;

opts.tau = 0.029150794692582; 
opts.cont_steps = 1; 
opts.tol = 3.0e-5; 
opts.mu = 1.0e-3; 
opts.tolpcg = 1.0e-1; 
opts.reg = 1.0e-1;
opts.maxpdNCGitr = 15;
opts.prec_tol = 1.0e-1;
opts.problem  = 'tv';
opts.verbose  = 2;
opts.user_fcn = @(options,primal,dual,f,g,H,P,d) ...
                                   user_fcn_fig1(options,primal,f,g,H,P,d);

fig1a = figure(1);
fig1b = figure(2);

xDC = pdNCGinterface(opts);

figure(fig1a);

leg = legend('$\lambda(\hat{B}) \ \ $'); 
set(leg,'FontSize',25,'Location','NorthEast','Interpreter','Latex'); 
                                         
xlabel('\bf{number of CG calls}', 'FontSize', 16);
ylabel('\bf{magnitude of spectrum}', 'FontSize', 16, 'Rotation', 90); 


figure(fig1b);

leg = legend('$\lambda(\tilde{N}^{-1}\hat{B}) \ \ $'); 
set(leg,'FontSize',25,'Location','NorthEast','Interpreter','Latex'); 
                                         
xlabel('\bf{number of PCG calls}', 'FontSize', 16);
ylabel('\bf{magnitude of spectrum}', 'FontSize', 16, 'Rotation', 90); 

figure(3);
plot([1:counter-1],cgiter_prec(1:counter-1),'b-o','LineWidth',4);
hold on;
plot([1:counter-1],cgiter_unprec(1:counter-1),'r--','LineWidth',4);

leg = legend('PCG','CG'); 
set(leg,'FontSize',20,'Location','NorthWest'); 
                                         
xlabel('\bf{number of CG/PCG calls}', 'FontSize', 16);
ylabel('\bf{Number of CG/PCG iterations}', 'FontSize', 16, 'Rotation', 90); 

figure(4);
plot([1:counter-1],cpu_prec(1:counter-1),'b-o','LineWidth',4);
hold on;
plot([1:counter-1],cpu_unprec(1:counter-1),'r--','LineWidth',4);

leg = legend('PCG','CG'); 
set(leg,'FontSize',20,'Location','NorthWest'); 
                                         
xlabel('\bf{number of CG/PCG calls}', 'FontSize', 16);
ylabel('\bf{CG/PCG time (sec)}', 'FontSize', 16, 'Rotation', 90); 


%% Call pdNCG solver for Figures 5.1c and 5.1d.
global fig2a;
global fig2b;
cgiter_prec = zeros(N,1);
cgiter_unprec = zeros(N,1);
cpu_prec = zeros(N,1);
cpu_unprec = zeros(N,1);
counter = 1;

fig2a = figure(5);
fig2b = figure(6);

opts.mu = 1.0e-5; 
opts.user_fcn = @(options,primal,dual,f,g,H,P,d) ...
                                   user_fcn_fig2(options,primal,f,g,H,P,d);
                               
xDC = pdNCGinterface(opts);

figure(fig2a);

leg = legend('$\lambda(\hat{B}) \ \ $'); 
set(leg,'FontSize',25,'Location','NorthEast','Interpreter','Latex'); 
                                         
xlabel('\bf{number of CG calls}', 'FontSize', 16);
ylabel('\bf{magnitude of spectrum}', 'FontSize', 16, 'Rotation', 90); 

figure(fig2b);

leg = legend('$\lambda(\tilde{N}^{-1}\hat{B}) \ \ $'); 
set(leg,'FontSize',25,'Location','NorthEast','Interpreter','Latex'); 
                                         
xlabel('\bf{number of PCG calls}', 'FontSize', 16);
ylabel('\bf{magnitude of spectrum}', 'FontSize', 16, 'Rotation', 90); 

figure(7);
plot([1:counter-1],cgiter_prec(1:counter-1),'b-o','LineWidth',4);
hold on;
plot([1:counter-1],cgiter_unprec(1:counter-1),'r--','LineWidth',4);

leg = legend('PCG','CG'); 
set(leg,'FontSize',20,'Location','NorthWest'); 
                                         
xlabel('\bf{number of CG/PCG calls}', 'FontSize', 16);
ylabel('\bf{Number of CG/PCG iterations}', 'FontSize', 16, 'Rotation', 90); 

figure(8);
plot([1:counter-1],cpu_prec(1:counter-1),'b-o','LineWidth',4);
hold on;
plot([1:counter-1],cpu_unprec(1:counter-1),'r--','LineWidth',4);

leg = legend('PCG','CG'); 
set(leg,'FontSize',20,'Location','NorthWest'); 
                                         
xlabel('\bf{number of CG/PCG calls}', 'FontSize', 16);
ylabel('\bf{CG/PCG time (sec)}', 'FontSize', 16, 'Rotation', 90); 

fprintf('Elapsed CPU time for this file: %d sec. \n', cputime - tt);
clear;

