% This script reproduces the results in Figure 5.3 in [1]
%
% [1] I. Dassios, K. Fountoulakis and J. Gondzio, "A Preconditioner for a
%     Primal-Dual Newton Conjugate Gradients Method for Compressed Sensing 
%     Problems". Technical Report ERGO 14-021.
%
% The problem instance that is used has been copied from 
% TFOCS_paperExamplesOnly package, url: http://cvxr.com/tfocs/download/.
%
% Some comments of the authors of TFOCS_paperExamplesOnly package follow.
%{
    Example of the Random Modulator Pre-Integrator (RMPI)
    with a radar pulse and continuous wave interferer

    ** Signal Processing Toolbox is required **

    Also relies on some helper files, which are included in the directory
     helper_functions. These files are:
        PsiWff and PsiTransposeWff  (the Gabor dictionary code)
        pulseTrain      (for setting up a sample pulse window )

    Should take about 4 or 5 minutes to run BP and a similar time
    to run Dantzig, so about 10 minutes to run this whole file.

%}

%-- test for toolboxes
% clear awgn randsrc
%if ~license('test','communication_toolbox')
    awgn = @(x,snr,ignore) x + ...
        10^( (10*log10(sum(abs(x(:)).^2)/length(x(:))) - snr)/20 )*randn(size(x));
    randsrc = @(varargin) sign( rand(varargin{:}) -.5 );
%end

clear;
tt_file = cputime;

%-- setup signal
factor = 4;
N   = factor*2048;
FS  = 5e9;  % Nyquist rate
BW  = FS/2; % bandwidth if we sampled at Nyquist rate
dT  = 1/FS;
KHz = 1e3; MHz = 1e6; GHz = 1e9; ms = 1e-6; ns = 1e-9; ps = 1e-12;
T   = N*dT;
t   = (0:N-1)*dT;


randn('state',324324); rand('state',324234);
% -- Add a large pu;se "x1"
f1  = rand*BW;  % frequency
a1  = 1;        % amplitude
p1  = rand*2*pi;% phase
% -- window by a trapezoidal window
tArrive = factor*40*ns;
tRise   = factor*20*ns;  tFall = tRise; 
tDur    = factor*300*ns;
tRep    = 0;
% -- get a trapezoidal window:
envelope1 = pulseTrain( t, tArrive, -tRise, tDur, tFall, tRep );
% -- this is unrealistic, so smooth it a bit:
f_cutoff = .004*FS;  % location of the first zero in frequency
n = find( t < 2/2/f_cutoff ); n = n(end);
kernel = ones(n,1); kernel = kernel / sum(kernel);
envelope1 = conv( envelope1, kernel, 'same' )';
x1  = a1*envelope1.*(sin(2*pi*f1*t + p1 ))';  % signal

% Add a small pulse "x2"
f2  = rand*BW;
a2  = 1e-3*a1;
p2  = rand*2*pi;
s   = a2*sin(2*pi*f2*t + p2);
% -- window by a trapezoidal window
tArrive = factor*45*ns;
tRise   = factor*20*ns;  tFall = tRise; 
tDur    = factor*300*ns;
tRep    = 0;
% -- get a trapezoidal window:
envelope2 = pulseTrain( t, tArrive, -tRise, tDur, tFall, tRep );
% -- this is unrealistic, so smooth it a bit:
f_cutoff = .004*FS;  % location of the first zero in frequency
n = find( t < 2/2/f_cutoff ); n = n(end);
kernel = ones(n,1); kernel = kernel / sum(kernel);
envelope2 = conv( envelope2, kernel, 'same' )';
x2 = envelope2.*s';

% -- combine big and small signals:
x_exact = (x1 + x2);

%% measurements
randn('state',320424); rand('state',3247825);
F_ADC       = 50*MHz;
nChannels   = 8;    % so, equivalent to a 400 MHz sampler
nSamples    = round(FS/F_ADC);
nPeriods    = floor( N/nSamples);
M           = nPeriods*nChannels;
Phi         = zeros(M,N);
for j = 0:nPeriods-1
    Phi(   j*nChannels + (1:nChannels) , j*nSamples + (1:nSamples)  ) = ...
        randsrc(nChannels,nSamples);
end
% add a bit to the final period
Phi( j*nChannels + (1:nChannels), nPeriods*nSamples:N ) = ...
    randsrc( nChannels, N - nPeriods*nSamples+1 );
normA2  = norm(Phi*Phi');
Phi = Phi/sqrt(normA2);

A   = @(x) Phi*x;
At  = @(y) Phi'*y;
b_exact = A(x_exact);
SNR = 60;   % in dB
x_noisy = awgn(x_exact, SNR, 'measured' );
z = x_exact - x_noisy;
b       = A(x_noisy);
EPS = norm(b-b_exact);
normA2  = norm(Phi*Phi');
fprintf('EPS is %.3f\n', EPS )

win = hamming(N);
%clf; periodogram( x_noisy,win, [], FS );

rms = @(x) norm(x)/sqrt(length(x));
snr = @(sig,noise) 20*log10( rms(sig)/rms(noise) );

fprintf('SNR of signal is %.1f dB; SNR of big tone is %.1f dB; SNR of small tone is %.1f dB\n',...
    snr(x_exact,z), snr(x1,z), snr(x2,z) );

sigma = std( b - b_exact );


%% setup dictionary
% The PsiWFF and PsiTransposeWFF code is a Gabor frame
% (i.e. a short-time Fourier transform)
% written by Peter Stobbe
% 
% PsiWFF is the synthesis operator, and acts on coefficients
% PsiTransposeWFF is the analysis operator, and acts on signals
gMax = 0;
gLevels = 5;
tRedundancy = 1;
fRedundancy = 1;
gWindow = 'isine';

lString = {};

logN = log2(N);
psi_A = @(y) PsiWFF(y,gWindow,logN,logN-gLevels,logN+gMax,tRedundancy,fRedundancy);
psiT_A = @(x) PsiTransposeWFF(x,gWindow,logN,logN-gLevels,logN+gMax,tRedundancy,fRedundancy);

N_Gabor = length( psiT_A( ones(N,1) ) );

psi  = psi_A;
psiT = psiT_A;

x  = x_exact;
y  = psiT(x);
x2 = psi(y);
d  = length(y);
% The frame is tight, so the psuedo-inverse is just the transpose:
fprintf('Error in PSI(PSI^* x ) - x is %e; N = %d, d = %d, d/N = %.1f\n', ...
    norm(x-x2),N,d,d/N);

normW2 = 1;

% look at coefficients
fcn = @(x) sort(abs(x),'descend');
c = fcn( psiT(x) );

% estimate a power-law decay rate, x_{k} = a*k^p
% offset = round(.2*d);
offset = 1;
cBulk = c( offset:round(.9*d) );
K = length(cBulk);
k = (1:K);
a = [ones(K,1), log(1:K)'] \ log(cBulk);
p = a(2);
% hold all
% semilogy( offset-1+ k, exp(a(1))*k.^p );
lString{end+1} = sprintf('p = %.3f',p );
fprintf('  and p is %.3f\n', p );

% end
%legend(lString)


%--  RECONSTRUCT via dual conic smoothing
W  = psiT;
Wt = psi;
opts        = [];
opts.maxits = 2500;
opts.maxmin = -1;
opts.saddle = 1;
opts.printEvery = 150;
opts.tol    = 1e-1;
opts.errFcn = { @(l,f,x) norm(x-x_exact)/norm(x_exact) };
opts.errFcn{end+1} = @(l,f,x) norm(A(x)-b)-EPS;
  
solver  = @solver_AT;
xx      = Phi\b;
alpha   = norm( W(xx), 1 );
beta    = norm( xx )^2/2;
mu      = 1.9*(alpha/beta);
  
scaleW  = sqrt(normA2/normW2);
MU      = scaleW*mu;
opts.L0 = 2/MU*normA2;
opts.stopCrit   = 2; % legacy
  
lambda0     = [];
  
xPlug   = zeros(N,1);
xOld    = xPlug;

%% Call pdNCG solver for Figure 5.2.
global counter;
%global fig1a;
%global fig1b;
global cgiter_prec;
global cpu_prec;
cgiter_prec = zeros(N,1);
cpu_prec = zeros(N,1);
global cgiter_unprec;
global cpu_unprec;
cgiter_unprec = zeros(N,1);
cpu_unprec = zeros(N,1);
counter = 1;

W  = psiT;
Wt = psi;
    
WW = linop_handles([N_Gabor,N],W,Wt );

WpdNCG = A_operator(@(q) WW(q,2), @(q) WW(q,1));

opts.n   = N;
opts.b   = b;
opts.A   = Phi;
opts.W   = WpdNCG;
opts.tau = 2.4167e-05;
opts.mu  = 1.0e-5;
opts.cont_steps = 1; 
opts.tolpcg = 1.0e-1; 
opts.reg = 1.0e-1;
opts.maxpdNCGitr = 15;
opts.prec_tol = 1.0e-1;
opts.tol = 1.0e-6; 
opts.verbose  = 2;
opts.prec_itrs = 5;
opts.mxitrpcg = 600;
opts.problem = 'wavelet';
opts.user_fcn = @(options,primal,dual,f,g,H,P,d) ...
                                   user_fcn_l1analysis(options,primal,f,g,H,P,d);

fig1a = figure(1);
fig1b = figure(2);

xDC = pdNCGinterface(opts);

figure(1);
plot([1:counter-1],cgiter_prec(1:counter-1),'b-o','LineWidth',4);
hold on;

figure(2);
plot([1:counter-1],cpu_prec(1:counter-1),'b-o','LineWidth',4);
hold on;

%opts.prec_tol = 1.0e-1; % enable preconditioner for all iterations
%opts.which_prec = 0;
        
%fprintf('\n \n pdNCG is called with approximate preconditioning \n');
%[~, ~, out] = pdNCGinterface(opts);
%fprintf('CPU time for approximate preconditioner case: %d \n',out.time);

%fprintf('Elapsed CPU time for this file: %d sec. \n', cputime - tt);
%clear;

%global fig1a;
%global fig1b;
cgiter_prec = zeros(N,1);
cpu_prec = zeros(N,1);
cgiter_unprec = zeros(N,1);
cpu_unprec = zeros(N,1);
counter = 1;

opts.prec_itrs = 10;
xDC = pdNCGinterface(opts);

figure(2);
plot([1:counter-1],cpu_prec(1:counter-1),'g-sq','LineWidth',4);
hold on;
figure(1);
plot([1:counter-1],cgiter_prec(1:counter-1),'g-sq','LineWidth',4);
hold on;

cgiter_prec = zeros(N,1);
cpu_prec = zeros(N,1);
cgiter_unprec = zeros(N,1);
cpu_unprec = zeros(N,1);
counter = 1;

opts.prec_itrs = 20;
xDC = pdNCGinterface(opts);

figure(2);
plot([1:counter-1],cpu_prec(1:counter-1),'y-d','LineWidth',4);
hold on;
plot([1:counter-1],cpu_unprec(1:counter-1),'r--','LineWidth',4);
hold on;
figure(1);
plot([1:counter-1],cgiter_prec(1:counter-1),'y-d','LineWidth',4);
hold on;
plot([1:counter-1],cgiter_unprec(1:counter-1),'r--','LineWidth',4);


figure(2);
leg = legend('PCG_{5}','PCG_{10}','PCG_{20}','CG'); 
set(leg,'FontSize',20,'Location','NorthWest'); 
                                         
xlabel('\bf{number of CG/PCG calls}', 'FontSize', 16);
ylabel('\bf{CG/PCG time (sec)}', 'FontSize', 16, 'Rotation', 90); 

figure(1);
leg = legend('PCG_{5}','PCG_{10}','PCG_{20}','CG'); 
set(leg,'FontSize',20,'Location','NorthWest'); 
                                         
xlabel('\bf{number of CG/PCG calls}', 'FontSize', 16);
ylabel('\bf{Number of CG/PCG iterations}', 'FontSize', 16, 'Rotation', 90); 