% This script reproduces Figure 6.2 in [1]
%
% [1] I. Dassios, K. Fountoulakis and J. Gondzio, "A Preconditioner for a
%     Primal-Dual Newton Conjugate Gradients Method for Compressed Sensing 
%     Problems". Technical Report ERGO 14-021.
%
% ** Signal Processing Toolbox is required **
%
% This file should take about 450 sec. CPU time to be completed.

clear;
tt = cputime;

randn('state',245);
rand('state',245);

% Generate phantom image
psize = 256;

n1 = psize; 
n2 = n1;
N = n1*n2;
n = max(n1,n2);

x = phantom(n);

x = x(1:n1,1:n2);
x_original = x;
mat = @(x) reshape(x,n1,n2);
vec = @(x) x(:);

% Partial Fourier measurements.
M = round( N / 4 );
omega = randn(N,1);
[temp,omega] = sort(omega);
omega = sort( omega(1:M) );
downsample = @(x) x(omega);
SS.type = '()'; SS.subs{1} = omega; SS.subs{2} = ':';
upsample = @(x) subsasgn( zeros(N,size(x,2)),SS,x);

% take partial 2D DCT measurements.  We have Af(At) = identity (not
% vice-versa of course). Randomly permuted.
rp = randperm(N);
[~,rp_inv] = sort(rp);
rpF = @(x) x(rp);
rp_invF = @(x) x(rp_inv);
        
my_dct2 = @(x) dct(dct(x).').';
my_idct2= @(x) idct(idct(x).').';
Af = @(x) downsample(vec( my_dct2( mat(rpF(mat(x) )) ) ) );
At = @(y) vec( rp_invF( my_idct2( mat( upsample( y ) ) ) ) );


A = linop_handles([M,N], Af, At );

myAwgn = @(x,snr) x + ...
        10^( (10*log10(sum(abs(x(:)).^2)/length(x(:))) - snr)/20 )*randn(size(x));
    
b_original = Af(vec(x_original));
snr = 10;  % SNR in dB
x_noisy = myAwgn( x_original, snr );
b   = Af( vec(x_noisy) );
EPS = norm(b-b_original);

mu  = 1e-5*norm( linop_TV(x_original) ,Inf);

x_ref   = x_original;

sz              = A([],0);
M               = sz{2}(1);
N               = sz{1}(1);
[n1,n2]         = size(x_original);
norm_x_ref      = norm(x_ref,'fro');
er_ref          = @(x) norm(vec(x)-vec(x_ref))/norm_x_ref;

fprintf('\tA is %s, %d x %d, signal has SNR %d dB\n', 'partial Fourier', M, N, round(snr) );

%% Call pdNCG solver for Figure 6.2 (make optimal x_{c,mu}).
Apdncg  = A_operator(@(w) A(w,1), @(w) A(w,2));

opts.n1  = n;
opts.n2  = n;
opts.b   = b;
opts.A   = Apdncg;
opts.tau = 0.053853287594141;
opts.tol = 2.0e-5;
opts.problem  = 'tv';
opts.verbose  = 2;
opts.prec_tol = 1.0e-4;

xopt = pdNCGinterface(opts);

%% Call pdNCG solver for Figure 6.2 (no continuation + no preconditioning).
global errors_ncnp;
global cpus_ncnp;
global counter_ncnp
global tstart;

counter_ncnp = 1;
errors_ncnp = zeros(n^2,1);
cpus_ncnp   = zeros(n^2,1);

opts.tol = 9.0e-2;
opts.cont_steps = 1;
opts.prec_tol = 1.0e-10;
opts.xopt = xopt;
opts.user_fcn = @(options,primal,dual,f,g,H,P,d) ...
                                 user_fcn_fig2_ncnp(options,primal,f,g,H,P,d);

tstart = cputime;
xDC = pdNCGinterface(opts);

%% Call pdNCG solver for Figure 6.2 (no continuation + preconditioning).
global errors_p;
global cpus_p;
global counter_p
global tstart;

counter_p = 1;
errors_p = zeros(n^2,1);
cpus_p   = zeros(n^2,1);

opts.tol = 5.0e-1;
opts.cont_steps = 1;
opts.prec_tol = 1.0e-1;
opts.xopt = xopt;
opts.user_fcn = @(options,primal,dual,f,g,H,P,d) ...
                                 user_fcn_fig2_p(options,primal,f,g,H,P,d);

tstart = cputime;
xDC = pdNCGinterface(opts);

%% Call pdNCG solver for Figure 6.2 (continuation + preconditioning).
global errors_cfp;
global cpus_cfp;
global counter_cfp
global tstart;

counter_cfp = 1;
errors_cfp = zeros(n^2,1);
cpus_cfp   = zeros(n^2,1);

opts.cont_steps = 5;
opts.prec_tol = 1.0e-1;
opts.user_fcn = @(options,primal,dual,f,g,H,P,d) ...
                                 user_fcn_fig2_cfp(options,primal,f,g,H,P,d);

tstart = cputime;
xDC = pdNCGinterface(opts);

%% Call pdNCG solver for Figure 6.2 (continuation + preconditioning at later stages).
global errors_chp;
global cpus_chp;
global counter_chp
global tstart;

counter_chp = 1;
errors_chp = zeros(n^2,1);
cpus_chp   = zeros(n^2,1);

opts.cont_steps = 5;
opts.prec_tol = 1.0e-4;
opts.user_fcn = @(options,primal,dual,f,g,H,P,d) ...
                                 user_fcn_fig2_chp(options,primal,f,g,H,P,d);

tstart = cputime;
xDC = pdNCGinterface(opts);

% Plot Figure 2
figure(1);
semilogx(cpus_ncnp(1:counter_ncnp-1),errors_ncnp(1:counter_ncnp-1),'black','LineWidth', 3);
hold on;
semilogx(cpus_p(1:counter_p-1),errors_p(1:counter_p-1),'b:sq','LineWidth', 4);
hold on;
semilogx(cpus_cfp(1:counter_cfp-1),errors_cfp(1:counter_cfp-1),'g-o','LineWidth', 3);
hold on;
semilogx(cpus_chp(1:counter_chp-1),errors_chp(1:counter_chp-1),'r--','LineWidth', 3);

axis tight;

leg = legend('no continuation, no prec.','only preconditioning','continuation and prec.','cont. and prec. at later stages');
set(leg,'FontSize',16,'Location','NorthEast');

xlabel('\bf{CPU time}', 'FontSize', 16);
ylabel('\bf{relative error}', 'FontSize', 16, 'Rotation', 90);

fprintf('Elapsed CPU time for this file: %d sec. \n', cputime - tt);
clear;
