% This script reproduces Figure 7.4 in [1]
%
% [1] I. Dassios, K. Fountoulakis and J. Gondzio, "A Preconditioner for a
%     Primal-Dual Newton Conjugate Gradients Method for Compressed Sensing 
%     Problems". Technical Report ERGO 14-021.
%
% The problem instance that is used has been copied from 
% TFOCS_paperExamplesOnly package, url: http://cvxr.com/tfocs/download/.
%
% Some comments of the authors of TFOCS_paperExamplesOnly package follow.
%{
    Example of the Random Modulator Pre-Integrator (RMPI)
    with a radar pulse and continuous wave interferer

    ** Signal Processing Toolbox is required **

    Also relies on some helper files, which are included in the directory
     helper_functions. These files are:
        PsiWff and PsiTransposeWff  (the Gabor dictionary code)
        pulseTrain      (for setting up a sample pulse window )

    Should take about 4 or 5 minutes to run BP and a similar time
    to run Dantzig, so about 10 minutes to run this whole file.

%}

%-- test for toolboxes
% clear awgn randsrc
%if ~license('test','communication_toolbox')
    awgn = @(x,snr,ignore) x + ...
        10^( (10*log10(sum(abs(x(:)).^2)/length(x(:))) - snr)/20 )*randn(size(x));
    randsrc = @(varargin) sign( rand(varargin{:}) -.5 );
%end

clear;
tt_file = cputime;

%-- setup signal
factor = 16;
N   = factor*2048;
FS  = 5e9;  % Nyquist rate
BW  = FS/2; % bandwidth if we sampled at Nyquist rate
dT  = 1/FS;
KHz = 1e3; MHz = 1e6; GHz = 1e9; ms = 1e-6; ns = 1e-9; ps = 1e-12;
T   = N*dT;
t   = (0:N-1)*dT;


randn('state',324324); rand('state',324234);
% -- Add a large pu;se "x1"
f1  = rand*BW;  % frequency
a1  = 1;        % amplitude
p1  = rand*2*pi;% phase
% -- window by a trapezoidal window
tArrive = factor*40*ns;
tRise   = factor*20*ns;  tFall = tRise; 
tDur    = factor*300*ns;
tRep    = 0;
% -- get a trapezoidal window:
envelope1 = pulseTrain( t, tArrive, -tRise, tDur, tFall, tRep );
% -- this is unrealistic, so smooth it a bit:
f_cutoff = .004*FS;  % location of the first zero in frequency
n = find( t < 2/2/f_cutoff ); n = n(end);
kernel = ones(n,1); kernel = kernel / sum(kernel);
envelope1 = conv( envelope1, kernel, 'same' )';
x1  = a1*envelope1.*(sin(2*pi*f1*t + p1 ))';  % signal

% Add a small pulse "x2"
f2  = rand*BW;
a2  = 1e-3*a1;
p2  = rand*2*pi;
s   = a2*sin(2*pi*f2*t + p2);
% -- window by a trapezoidal window
tArrive = factor*45*ns;
tRise   = factor*20*ns;  tFall = tRise; 
tDur    = factor*300*ns;
tRep    = 0;
% -- get a trapezoidal window:
envelope2 = pulseTrain( t, tArrive, -tRise, tDur, tFall, tRep );
% -- this is unrealistic, so smooth it a bit:
f_cutoff = .004*FS;  % location of the first zero in frequency
n = find( t < 2/2/f_cutoff ); n = n(end);
kernel = ones(n,1); kernel = kernel / sum(kernel);
envelope2 = conv( envelope2, kernel, 'same' )';
x2 = envelope2.*s';

% -- combine big and small signals:
x_exact = (x1 + x2);

%% measurements
randn('state',320424); rand('state',3247825);
F_ADC       = 50*MHz;
nChannels   = 8;    % so, equivalent to a 400 MHz sampler
nSamples    = round(FS/F_ADC);
nPeriods    = floor( N/nSamples);
M           = nPeriods*nChannels;
Phi         = zeros(M,N);
for j = 0:nPeriods-1
    Phi(   j*nChannels + (1:nChannels) , j*nSamples + (1:nSamples)  ) = ...
        randsrc(nChannels,nSamples);
end
% add a bit to the final period
Phi( j*nChannels + (1:nChannels), nPeriods*nSamples:N ) = ...
    randsrc( nChannels, N - nPeriods*nSamples+1 );
normA2  = norm(Phi*Phi');
Phi = Phi/sqrt(normA2);

A   = @(x) Phi*x;
At  = @(y) Phi'*y;
b_exact = A(x_exact);
SNR = 60;   % in dB
x_noisy = awgn(x_exact, SNR, 'measured' );
z = x_exact - x_noisy;
b       = A(x_noisy);
EPS = norm(b-b_exact);
normA2  = norm(Phi*Phi');
fprintf('EPS is %.3f\n', EPS )

win = hamming(N);
clf; periodogram( x_noisy,win, [], FS );

rms = @(x) norm(x)/sqrt(length(x));
snr = @(sig,noise) 20*log10( rms(sig)/rms(noise) );

fprintf('SNR of signal is %.1f dB; SNR of big tone is %.1f dB; SNR of small tone is %.1f dB\n',...
    snr(x_exact,z), snr(x1,z), snr(x2,z) );

sigma = std( b - b_exact );


%% setup dictionary
% The PsiWFF and PsiTransposeWFF code is a Gabor frame
% (i.e. a short-time Fourier transform)
% written by Peter Stobbe
% 
% PsiWFF is the synthesis operator, and acts on coefficients
% PsiTransposeWFF is the analysis operator, and acts on signals
gMax = 0;
gLevels = 5;
tRedundancy = 1;
fRedundancy = 1;
gWindow = 'isine';

lString = {};

logN = log2(N);
psi_A = @(y) PsiWFF(y,gWindow,logN,logN-gLevels,logN+gMax,tRedundancy,fRedundancy);
psiT_A = @(x) PsiTransposeWFF(x,gWindow,logN,logN-gLevels,logN+gMax,tRedundancy,fRedundancy);

N_Gabor = length( psiT_A( ones(N,1) ) );

psi  = psi_A;
psiT = psiT_A;

x  = x_exact;
y  = psiT(x);
x2 = psi(y);
d  = length(y);
% The frame is tight, so the psuedo-inverse is just the transpose:
fprintf('Error in PSI(PSI^* x ) - x is %e; N = %d, d = %d, d/N = %.1f\n', ...
    norm(x-x2),N,d,d/N);

normW2 = 1;

% look at coefficients
fcn = @(x) sort(abs(x),'descend');
c = fcn( psiT(x) );

% estimate a power-law decay rate, x_{k} = a*k^p
% offset = round(.2*d);
offset = 1;
cBulk = c( offset:round(.9*d) );
K = length(cBulk);
k = (1:K);
a = [ones(K,1), log(1:K)'] \ log(cBulk);
p = a(2);
% hold all
% semilogy( offset-1+ k, exp(a(1))*k.^p );
lString{end+1} = sprintf('p = %.3f',p );
fprintf('  and p is %.3f\n', p );

% end
legend(lString)


%--  RECONSTRUCT via dual conic smoothing
W  = psiT;
Wt = psi;
opts        = [];
opts.maxits = 2500;
opts.maxmin = -1;
opts.saddle = 1;
opts.printEvery = 150;
opts.tol    = 1e-1;
opts.errFcn = { @(l,f,x) measure_snr(x,x_exact) };
opts.errFcn{end+1} = @(l,f,x) norm(A(x)-b)-EPS;
  
solver  = @solver_AT;
xx      = Phi\b;
alpha   = norm( W(xx), 1 );
beta    = norm( xx )^2/2;
mu      = 1.9*(alpha/beta);
  
scaleW  = sqrt(normA2/normW2);
MU      = scaleW*mu;
opts.L0 = 2/MU*normA2;
opts.stopCrit   = 2; % legacy
  
lambda0     = [];
  
xPlug   = zeros(N,1);
xOld    = xPlug;
  
% Here are the relevant parameters
tol_continuation    = 1e-3;
tol_increase        = 1.5;

WW = linop_handles([N_Gabor,N],W,Wt );
    
opts.tol    = 1e-1;
iter_time = cputime;
opts.errFcn{end+1} = @(l,f,primal) cputime - iter_time; 

snr_tfocs = zeros(3*length(xPlug),1);
times_tfocs = zeros(3*length(xPlug),1);
counter_tfocs = 1;

k_end   = 12;
tt = cputime;
for k = 1:k_end  % continuation loop       
        
    opts.normA2     = normA2;
    opts.normW2     = normW2;
    [x,out,optsOut] = solver_sBPDN_W(Phi,WW,b,EPS,mu,xPlug,lambda0,opts );
    snr_tfocs(counter_tfocs:counter_tfocs+out.niter-1) = out.err(1:out.niter,1);
    times_tfocs(counter_tfocs:counter_tfocs+out.niter-1) = out.err(1:out.niter,3);
    counter_tfocs = counter_tfocs + out.niter;
    
    lambda = out.dual;
    fprintf('Relative error is %.2e\n', norm(x-x_exact)/norm(x_exact) );
        
    % continuation update
    xPlug   = x + k/(k+3)*( x - xOld );
    lambda0 = lambda;
        
    opts.tol = opts.tol/tol_increase;
    if k > 1
      change = norm( x - xOld )/norm(x);
      fprintf('After continuation %d, rel. change in x is %.2e\n',k,change);
      if change < tol_continuation, disp('Change is small, so exiting from continuation'); break; end
    end
    xOld    = x;
end
t_TFOCS = cputime - tt;

fprintf('TOTAL ELAPSED CPU TIME IS %d \n', t_TFOCS );

%-- Make plot
figure(1); clf;
win = hamming(N);
NN = 1*N;
  
[Pxx,F] = periodogram( x_exact, win, NN, FS );
Pxx = db(Pxx,'power');
F = F/GHz;
plot( F, Pxx,'k','linewidth',2);
hold all
  
[Pxx,F] = periodogram( x, win, NN, FS );
Pxx = db(Pxx,'power');
F = F/GHz;
plot( F, Pxx,'r.','markersize',20 );
  
xlabel('Frequency (GHz)','fontsize',20);
ylabel('Power/frequency (dB/Hz)','fontsize',20);
  
ylim([-220,-50])
  
h = legend('Original signal','Recovered signal');
set(h,'fontsize',20 );
%title('TFOCS');

[psnr_val, snr] = psnr(x, x_exact);
        SNR_xDC = fix(snr*10)/10;
fprintf('TFOCS con, SNR is %.1f dB, CPU time is %.1f \n \n',...
        SNR_xDC, t_TFOCS);


%% pdNCG.
W  = psiT;
Wt = psi;
    
WW = linop_handles([N_Gabor,N],W,Wt );

WpdNCG = A_operator(@(q) WW(q,2), @(q) WW(q,1));

opts.n   = N;
opts.b   = b;
opts.A   = Phi;
opts.W   = WpdNCG;
opts.tau = 0.00695*2*norm(out.dual{2},inf)/norm(out.dual{1});
opts.mu  = 1.0e-6;
opts.problem = 'wavelet';

global snr_pdncg;
snr_pdncg = zeros(3*length(xPlug),1);
global times_pdncg;
times_pdncg   = zeros(3*length(xPlug),1);
global counter_pdncg;
counter_pdncg = 1;
global tstart;
opts.user_fcn = @(options,primal,dual,f,g,H,P,d) measure_snr_pdncg(primal,x_exact);


tt = cputime;
tstart = tt;
tic;

% Check input arguments.
opts = checkOptions(opts);

% initialize data.
order_mu  = max(-floor(log(abs(opts.mu))./log(10)),0);
order_tau = max(-floor(log(abs(opts.tau))./log(10)),0);
    
t_mu  = max(order_mu-1,0);
t_tau = max(order_tau-1,0);

if t_mu == 0
    mus_vector = ones(opts.cont_steps,1).*opts.mu;
else
    mus_vector = logspace(log10(0.1),log10(opts.mu),opts.cont_steps);
end

if t_tau == 0
    taus_vector = ones(opts.cont_steps,1).*opts.tau;
else
    taus_vector = logspace(log10(0.1),log10(opts.tau),opts.cont_steps);
end

opts.tol = 1.0e-3;

% Begin manual continuation.
for i=1:opts.cont_steps
    
    opts.mu  = mus_vector(i);  
    opts.tau = taus_vector(i);
    
    fprintf('%5s:  %d \n','Total continuation steps',opts.cont_steps);
    fprintf('%5s: %d \n','Current continuation step',i);   
    fprintf('%5s: %d \n','Current mu ',opts.mu);
    fprintf('%5s: %d \n','Current tau',opts.tau); 
    
    [xDC, y, out] = pdNCG_W(opts);
    opts.x0 = xDC; 
    opts.y0 = y;
    opts.tol = max(opts.tol/10,1.0e-8);
    
end
time_pdncg = cputime - tt;
    
fprintf('TOTAL ELAPSED TIME IS %d\n',time_pdncg );

%-- Make plot
figure(2); clf;
win = hamming(N);
NN = 1*N;
  
[Pxx,F] = periodogram( x_exact, win, NN, FS );
Pxx = db(Pxx,'power');
F = F/GHz;
plot( F, Pxx,'k','linewidth',2);
hold all
  
[Pxx,F] = periodogram( xDC, win, NN, FS );
Pxx = db(Pxx,'power');
F = F/GHz;
plot( F, Pxx,'r.','markersize',20 );
  
xlabel('Frequency (GHz)','fontsize',20);
ylabel('Power/frequency (dB/Hz)','fontsize',20);
  
ylim([-220,-50])
  
h = legend('Original signal','Recovered signal');
set(h,'fontsize',20 );
%title('pdNCG');
  
[psnr_val, snr] = psnr(xDC, x_exact);
        SNR_xDC = fix(snr*10)/10;
fprintf('pdNCG, SNR is %.1f dB, CPU time is %.1f \n \n',...
        SNR_xDC, time_pdncg);

    
figure(3);
plot(times_pdncg(1:1:counter_pdncg-1),snr_pdncg(1:counter_pdncg-1),'b','LineWidth', 3);
hold on;
plot(times_tfocs(1:counter_tfocs-1),snr_tfocs(1:counter_tfocs-1),'r--','LineWidth', 3);

axis tight;

leg = legend('pdNCG','TFOCS\_con');
set(leg,'FontSize',16,'Location','SouthEast');

xlabel('\bf{CPU time}', 'FontSize', 16);
ylabel('\bf{SNR}', 'FontSize', 16, 'Rotation', 90);    
    
fprintf('Elapsed CPU time for this file: %d sec. \n', cputime - tt_file);

clear;
